from .deployment import Deployment
from .repository import Repository

from .client_options import ClientConfig

from .create_deployment import CreateDeployment

from .workspace import Workspace

from .deploy_options import DeployOptions

from .prediction import V1Prediction, V2Prediction



import base64
import logging
from typing import Optional, List, Any
import os

import requests
from pydantic import parse_obj_as

from deeploy.models import Deployment, Repository, CreateDeployment, Workspace, V1Prediction, V2Prediction
from deeploy.enums import ModelType, ExplainerType, PredictionVersion


class DeeployService(object):
    """ 
    A class for interacting with the Deeploy API
    """

    def __init__(self, access_key: str, secret_key: str, host: str, insecure=False) -> None:
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__host = 'http://api.%s' % host if insecure else 'https://api.%s' % host

        if not self.__keys_are_valid():
            raise Exception('Access keys are not valid')
        return

    def __keys_are_valid(self) -> bool:
        host_for_testing = '%s/v2/workspaces' % self.__host
        workspaces_response = requests.get(
            host_for_testing, auth=(self.__access_key, self.__secret_key))
        if self.__request_is_successful(workspaces_response):
            return True
        return False

    def __request_is_successful(self, request: requests.Response) -> bool:
        if str(request.status_code)[0] == '2':
            return True
        return False

    def get_repositories(self, workspace_id: str) -> List[Repository]:
        url = '%s/v2/workspaces/%s/repositories' % (
            self.__host, workspace_id)
        params = {
            'isArchived': False,
        }

        repositories_response = requests.get(
            url, params=params, auth=(self.__access_key, self.__secret_key))

        repositories = parse_obj_as(
            List[Repository], repositories_response.json())

        return repositories

    def get_repository(self, workspace_id: str, repository_id: str) -> Repository:
        url = '%s/v2/workspaces/%s/repositories/%s' % (
            self.__host, workspace_id, repository_id)

        repository_response = requests.get(
            url, auth=(self.__access_key, self.__secret_key))
        if not self.__request_is_successful(repository_response):
            raise Exception('Repository does not exist in the workspace.')

        repository = parse_obj_as(Repository, repository_response.json())

        return repository

    def create_deployment(self, workspace_id: str, deployment: CreateDeployment) -> Deployment:
        url = '%s/v2/workspaces/%s/deployments' % (self.__host, workspace_id)
        data = deployment.to_request_body()

        deployment_response = requests.post(
            url, json=data, auth=(self.__access_key, self.__secret_key))
        if not self.__request_is_successful(deployment_response):
            raise Exception('Failed to create the deployment: %s' % str(deployment_response.json()) )

        deployment = parse_obj_as(
            Deployment, deployment_response.json())

        return deployment

    def get_workspace(self, workspace_id: str) -> Workspace:
        url = '%s/v2/workspaces/%s' % (self.__host, workspace_id)

        workspace_response = requests.get(
            url, auth=(self.__access_key, self.__secret_key))
        if not self.__request_is_successful(workspace_response):
            raise Exception('Workspace does not exist.')

        workspace = parse_obj_as(Workspace, workspace_response.json())

        return workspace

    def upload_blob_file(self, local_file_path: str, relative_folder_path: str, workspace_id: str, repository_id: str, uuid: str) -> str:
        url = '%s/v2/workspaces/%s/repositories/%s/upload' % (self.__host, workspace_id, repository_id)
        params = {
            'commitSha': uuid,
            'folderPath': relative_folder_path,
        }
        files = { 'file': open(local_file_path, 'rb') }
        r = requests.post(url, files=files, params=params, auth=(self.__access_key, self.__secret_key))
        blob_storage_path = r.json()['data']['referencePath']
        return blob_storage_path

    def __check_prediction_version(self, prediction_response: dict) -> PredictionVersion:
        if 'predictions' in prediction_response.json():
             return PredictionVersion.V1
        else:
            return PredictionVersion.V2

    def __parse_prediction(self, prediction_response: dict) -> V1Prediction or V2Prediction:
        if self.__check_prediction_version(prediction_response) == PredictionVersion.V1:
             prediction = parse_obj_as(
                V1Prediction, prediction_response.json())
        else:
            prediction = parse_obj_as(
                V2Prediction, prediction_response.json())
        return prediction

    def predict(self, workspace_id: str, deployment_id: str, request_body: dict) -> V1Prediction or V2Prediction:
        url = '%s/v2/workspaces/%s/deployments/%s/predict' % (self.__host, workspace_id, deployment_id)
        params = {
            'skipLog': str(False).lower(),
        }

        prediction_response = requests.post(
            url, json=request_body, params=params, auth=(self.__access_key, self.__secret_key))
            
        if not self.__request_is_successful(prediction_response):
            raise Exception('Failed to call predictive model.')
        prediction = self.__parse_prediction(prediction_response)
        return prediction


    def explain(self, workspace_id: str, deployment_id: str, request_body: dict, image: bool = False) -> object:
        url = '%s/v2/workspaces/%s/deployments/%s/explain' % (self.__host, workspace_id, deployment_id)
        params = {
            'image': str(image).lower(),
            'skipLog': str(False).lower(),
            'raw': str(False).lower(),
        }

        explanation_reponse = requests.post(
            url, json=request_body, params=params, auth=(self.__access_key, self.__secret_key))
        if not self.__request_is_successful(explanation_reponse):
            raise Exception('Failed to call explainer model.')
        explanation = explanation_reponse.json()
        return explanation




from typing import List, Optional, Dict, Any

from pydantic import BaseModel

class Prediction(BaseModel):
    pass

class V1Prediction(Prediction):
    predictions: List[Any]

class V2Prediction(Prediction):
    id: str
    model_name: str
    model_version: Optional[str]
    parameters: Optional[Dict]
    outputs: List[dict]



from typing import Any

from deeploy.enums import ExplainerType


class BaseExplainer:

    def __init__(self, explainer_object: Any) -> None:
        return

    def save(self, local_folder_path: str) -> None:
        return

    def get_explainer_type(self) -> ExplainerType:
        return



from typing import Optional, List, Any

from pydantic import BaseModel


class DeployOptions(BaseModel):
    """Class that contains the options for deploying a model

    Attributes:
        name (str): name of the deployment
        model_serverless (bool, optional): whether to deploy the model in 
            a serverless fashion. Defaults to False
        explainer_serverless (bool, optional): whether to deploy the model in 
            a serverless fashion. Defaults to False
        description (str, optional): the description of the deployment
        example_input (List, optional): list of example input parameters for the model
        example_output (List, optional): list of example output for the model
        pytorch_model_file_path (str, optional): absolute or relative path to the .py file 
            containing the pytorch model class definition
        pytorch_torchserve_handler_name (str, optional): TorchServe handler name. One of 
            ['image_classifier', 'image_classifier', 'object_detector', 'text_classifier']. 
            See the (TorchServe documentation)[https://github.com/pytorch/serve/blob/master/docs/default_handlers.md#torchserve-default-inference-handlers] 
            for more info.
    """
    name: str
    model_serverless = False
    explainer_serverless = False
    description: Optional[str]
    example_input: Optional[List[Any]]
    example_output: Optional[List[Any]]
    pytorch_model_file_path: Optional[str]
    pytorch_torchserve_handler_name: Optional[str]



from typing import List, Optional

from pydantic import BaseModel

from deeploy.common import to_lower_camel


class Repository(BaseModel):
    id: str
    name: str
    status: int
    is_archived: bool
    workspace_id: str
    is_public: Optional[bool]
    git_ssh_pull_link: str
    created_at: str
    updated_at: str

    class Config:
        alias_generator = to_lower_camel



from typing import Optional, List, Any

from pydantic import BaseModel

from deeploy.common import to_lower_camel
from deeploy.enums import ModelType, ExplainerType


class Deployment(BaseModel):
    name: str
    versions: Optional[List[Any]]
    workspace_id: str
    description: Optional[str]
    example_input: Optional[List[Any]]
    example_output: Optional[List[Any]]
    status: int
    owner_id: str
    kf_serving_id: Optional[str]
    public_url: Optional[str]
    id: str
    created_at: str
    updated_at: str

    class Config:
        alias_generator = to_lower_camel


class DeployOptions(BaseModel):
    """
    Class that contains the options for deploying a model

    Attributes:
      name: name of the deployment
      model_serverless: boolean indicating whether to deploy the model in 
        a serverless fashion. Defaults to False
      explainer_serverless: boolean indicating whether to deploy the model in 
        a serverless fashion. Defaults to False
      description: string with the description of the deployment
      example_input: list of example input parameters for the model
      example_output: list of example output for the model
    """
    name: str
    model_serverless = False
    explainer_serverless = False
    description: Optional[str]
    example_input: Optional[List[Any]]
    example_output: Optional[List[Any]]



from typing import Any
from os.path import join

from sklearn.base import BaseEstimator
from joblib import dump

from . import BaseModel
from deeploy.enums import ModelType


class SKLearnModel(BaseModel):

    __sklearn_model: BaseEstimator

    def __init__(self, model_object: Any, **kwargs) -> None:

        if not issubclass(type(model_object), BaseEstimator):
            raise Exception('Not a valid SKLearn class')

        self.__sklearn_model = model_object
        return

    def save(self, local_folder_path: str) -> None:
        dump(self.__sklearn_model, join(local_folder_path, 'model.joblib'))
        return

    def get_model_type(self) -> ModelType:
        return ModelType.SKLEARN



from typing import List, Optional

from pydantic import BaseModel

from deeploy.common import to_lower_camel


class Workspace(BaseModel):
    id: str
    name: str
    description: Optional[str]
    owner_id: str
    created_at: str
    updated_at: str

    class Config:
        alias_generator = to_lower_camel



from typing import Any
from os.path import join, exists
import os
from shutil import copy, move
import random
import string
import subprocess

from torch.nn import Module
from torch import save

from . import BaseModel
from deeploy.enums import ModelType
from deeploy.common import PYTORCH_CONFIG_FILE


class PyTorchModel(BaseModel):

    __pytorch_model: Module
    __model_file_path: str
    __handler_file_path: str = None

    def __init__(self, model_object: Any, pytorch_model_file_path: str, pytorch_torchserve_handler_name: str = 'image_classifier', **kwargs) -> None:

        if not issubclass(type(model_object), Module):
            raise Exception('Not a valid PyTorch class')

        if not exists(pytorch_model_file_path):
            raise Exception('The Pytorch model file does not exist')

        if handler_file_path:
            if not exists(pytorch_model_file_path):
                raise Exception('The Pytorch model file does not exist')

        if not (pytorch_model_file_path.endswith('.py') or pytorch_model_file_path.endswith('.ipynb')):
            raise Exception(
                'The Pytorch model file is not a supported file type. Use .py or .ipynb')

        self.__pytorch_model = model_object
        self.__model_file_path = pytorch_model_file_path
        self.__handler_name = handler_name
        return

    def save(self, local_folder_path: str) -> None:
        serialized_model_path = join(os.getcwd(), 'model.pt')
        mar_folder_path = join(local_folder_path, 'model-store')
        save(self.__pytorch_model.state_dict(), serialized_model_path)

        # torch-model-archiver --model-name densenet161 --version 1.0 --model-file examples/image_classifier/densenet_161/model.py --serialized-file densenet161-8d451a50.pth --extra-files examples/image_classifier/index_to_name.json --handler image_classifier
        mar_command = "torch-model-archiver --model-name model --version 1.0 --serialized-file %s \
            --export-path %s" % (serialized_model_path, mar_folder_path)

        if self.__model_file_path.endswith('.py'):
            mar_command += " --model-file %s" % self.__model_file_path

        elif self.__model_file_path.endswith('.ipynb'):
            convert_command = "ipython nbconvert --to script %s" % self.__model_file_path

            ipy_process = subprocess.Popen(
                convert_command.split(), stdout=subprocess.PIPE)
            _, error = ipy_process.communicate()
            if error:
                raise Exception(error)

            mar_command += " --model-file %s" % self.__model_file_path.replace(
                '.ipynb', '.py')

        mar_command += " --handler %s" % self.__handler_name

        if not os.path.exists(mar_folder_path):
            os.makedirs(mar_folder_path)

        mar_process = subprocess.Popen(
            mar_command.split(), stdout=subprocess.PIPE)
        _, error = mar_process.communicate()
        if error:
            raise Exception(error)
        mar_process.wait()

        config_folder_path = join(local_folder_path, 'config')
        if not os.path.exists(config_folder_path):
            os.makedirs(config_folder_path)
        
        config_file_path = join(config_folder_path, 'config.properties')
        with open(config_file_path, "w+") as config_file:
            config_file.write(PYTORCH_CONFIG_FILE)

        return

    def get_model_type(self) -> ModelType:
        return ModelType.PYTORCH



import os
import shutil
import logging


def to_lower_camel(string: str) -> str:
    if string == 'git_ssh_pull_link':
        return 'gitSSHPullLink'
    if string == 'public_url':
        return 'publicURL'

    camel_case = ''.join(word.capitalize() for word in string.split('_'))
    lower_camel_case = camel_case[0].lower() + camel_case[1:]
    return lower_camel_case


def delete_all_contents_in_directory(folder_path: str) -> None:
    for filename in os.listdir(folder_path):
        file_path = os.path.join(folder_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            logging.error('Failed to delete %s. Reason: %s' % (file_path, e))
    return


def directory_exists(folder_path: str) -> bool:
    return os.path.exists(folder_path)

def directory_empty(folder_path: str) -> bool:
    return len(os.listdir(folder_path)) == 0



from onnx import ModelProto

from . import BaseModel
from deeploy.enums import ModelType


class ONNXModel(BaseModel):
    # TODO
    def save(self, local_folder_path: str) -> None:
        return

    def get_model_type(self) -> ModelType:
        return ModelType.ONNX


from typing import Any
from os.path import join

from xgboost import XGBClassifier, Booster
from joblib import dump

from . import BaseModel
from deeploy.enums import ModelType


class XGBoostModel(BaseModel):

    __xgboost_model: XGBClassifier

    def __init__(self, model_object: Any, **kwargs) -> None:

        if not issubclass(type(model_object), XGBClassifier) and \
            not issubclass(type(model_object), Booster):
            raise Exception('Not a valid XGBoost class')

        self.__xgboost_model = model_object
        return

    def save(self, local_folder_path: str) -> None:
        self.__xgboost_model.save_model(join(local_folder_path, 'model.bst'))
        return

    def get_model_type(self) -> ModelType:
        return ModelType.XGBOOST



from typing import Any, List
import inspect

from deeploy.enums import ModelType
from deeploy.services.models import BaseModel


class ModelWrapper:

    __model_helper: BaseModel

    def __init__(self, model_object: Any, **kwargs) -> None:

        self.__model_helper = self.__get_model_helper(model_object, **kwargs)

        return

    def save(self, local_folder_path: str) -> None:
        self.__model_helper.save(local_folder_path)
        return

    def get_model_type(self) -> ModelType:
        return self.__model_helper.get_model_type()

    def __get_model_type(self, model_object: Any) -> ModelType:

        base_classes = list(map(lambda x: x.__module__ + '.' +
                                x.__name__, inspect.getmro(type(model_object))))

        if self.__is_sklearn(base_classes):
            return ModelType.SKLEARN
        if self.__is_xgboost(base_classes):
            return ModelType.XGBOOST
        if self.__is_pytorch(base_classes):
            return ModelType.PYTORCH
        if self.__is_tensorflow(base_classes):
            return ModelType.TENSORFLOW
        if self.__is_triton(base_classes):
            return ModelType.TRITON
        if self.__is_onnx(base_classes):
            return ModelType.ONNX

        raise NotImplementedError(
            'This model type is not implemented by Deeploy')

    def __get_model_helper(self, model_object, **kwargs) -> BaseModel:

        model_type = self.__get_model_type(model_object)

        # only import the helper class when it is needed
        if model_type == ModelType.SKLEARN:
            from deeploy.services.models.sklearn import SKLearnModel
            return SKLearnModel(model_object, **kwargs)
        if model_type == ModelType.XGBOOST:
            from deeploy.services.models.xgboost import XGBoostModel
            return XGBoostModel(model_object, **kwargs)
        if model_type == ModelType.PYTORCH:
            from deeploy.services.models.pytorch import PyTorchModel
            return PyTorchModel(model_object, **kwargs)
        if model_type == ModelType.TENSORFLOW:
            from deeploy.services.models.tensorflow import TensorFlowModel
            return TensorFlowModel(model_object, **kwargs)
        if model_type == ModelType.TRITON:
            from deeploy.services.models.triton import TritonModel
            return TritonModel(model_object, **kwargs)
        if model_type == ModelType.ONNX:
            from deeploy.services.models.onnx import ONNXModel
            return ONNXModel(model_object, **kwargs)

    def __is_sklearn(self, base_classes: List[str]) -> bool:
        return 'sklearn.base.BaseEstimator' in base_classes and not 'xgboost.sklearn.XGBModel' in base_classes

    def __is_xgboost(self, base_classes: List[str]) -> bool:
        return 'xgboost.sklearn.XGBModel' in base_classes or 'xgboost.core.Booster' in base_classes

    def __is_pytorch(self, base_classes: List[str]) -> bool:
        return 'torch.nn.modules.module.Module' in base_classes

    def __is_tensorflow(self, base_classes: List[str]) -> bool:
        return 'tensorflow.python.module.module.Module' in base_classes

    def __is_onnx(self, model: Any) -> bool:
        # TODO
        return False

    def __is_triton(self, model: Any) -> bool:
        # TODO
        return False



from typing import Any
from os.path import join
import inspect

import dill
from alibi.api.interfaces import Explainer

from . import BaseExplainer
from deeploy.enums import ExplainerType


class AlibiExplainer(BaseExplainer):

    __alibi_explainer: Explainer

    def __init__(self, explainer_object: Any) -> None:

        if not issubclass(type(explainer_object), Explainer):
            raise Exception('Not a valid Alibi class')

        self.__alibi_explainer = explainer_object
        return

    def save(self, local_folder_path: str) -> None:
        with open(join(local_folder_path, 'explainer.dill'), 'wb') as f:
            dill.dump(self.__alibi_explainer, f)
        return

    def get_explainer_type(self) -> ExplainerType:
        base_classes = list(map(lambda x: x.__module__ + '.' +
                                x.__name__, inspect.getmro(type(self.__alibi_explainer))))
                                
        if 'alibi.explainers.anchor_text.AnchorText' in base_classes:
            return ExplainerType.ANCHOR_TEXT
        if 'alibi.explainers.anchor_image.AnchorImage' in base_classes:
            return ExplainerType.ANCHOR_IMAGES
        if 'alibi.explainers.anchor_tabular.AnchorTabular' in base_classes:
            return ExplainerType.ANCHOR_TABULAR



from typing import Any, List
import inspect

from deeploy.enums import ExplainerType
from deeploy.services.explainers import BaseExplainer


class ExplainerWrapper:

    __explainer_helper: BaseExplainer

    def __init__(self, explainer_object: Any) -> None:

        self.__explainer_helper = self.__get_explainer_helper(explainer_object)
        return

    def save(self, local_folder_path: str) -> None:
        self.__explainer_helper.save(local_folder_path)
        return

    def get_explainer_type(self) -> ExplainerType:
        return self.__explainer_helper.get_explainer_type()

    def __get_explainer_type(self, model_object: Any) -> ExplainerType:

        base_classes = list(map(lambda x: x.__module__ + '.' +
                                x.__name__, inspect.getmro(type(model_object))))

        if self.__is_alibi_anchor_text(base_classes):
            return ExplainerType.ANCHOR_TEXT
        if self.__is_alibi_anchor_images(base_classes):
            return ExplainerType.ANCHOR_IMAGES
        if self.__is_alibi_anchor_tabular(base_classes):
            return ExplainerType.ANCHOR_TABULAR
        if self.__is_shap_kernel(base_classes):
            return ExplainerType.SHAP_KERNEL

        raise NotImplementedError(
            'This explainer type is not implemented by Deeploy')

    def __get_explainer_helper(self, explainer_object) -> BaseExplainer:

        explainer_type = self.__get_explainer_type(explainer_object)

        # only import the helper class when it is needed
        if explainer_type == ExplainerType.ANCHOR_TEXT or \
                explainer_type == ExplainerType.ANCHOR_IMAGES or \
                explainer_type == ExplainerType.ANCHOR_TABULAR:
            from deeploy.services.explainers.alibi import AlibiExplainer
            return AlibiExplainer(explainer_object)
        if explainer_type == ExplainerType.SHAP_KERNEL:
            from deeploy.services.explainers.shap import SHAPExplainer
            return SHAPExplainer(explainer_object)

    def __is_alibi_anchor_text(self, base_classes: List[str]) -> bool:
        return 'alibi.explainers.anchor_text.AnchorText' in base_classes

    def __is_alibi_anchor_images(self, base_classes: List[str]) -> bool:
        return 'alibi.explainers.anchor_image.AnchorImage' in base_classes

    def __is_alibi_anchor_tabular(self, base_classes: List[str]) -> bool:
        return 'alibi.explainers.anchor_tabular.AnchorTabular' in base_classes

    def __is_shap_kernel(self, base_classes: List[str]) -> bool:
        return 'shap.explainers.kernel.KernelExplainer' in base_classes



from typing import Optional, List, Any

from pydantic import BaseModel

from deeploy.enums import ModelType, ExplainerType


class CreateDeployment(BaseModel):
    """
    """
    repository_id: str
    name: str
    description: Optional[str]
    example_input: Optional[List[Any]]
    example_output: Optional[Any]
    model_type: ModelType
    model_serverless: Optional[bool] = False
    explainer_type: ExplainerType
    explainer_serverless: Optional[bool] = False
    branch_name: str
    commit: str

    def to_request_body(self):
        return {
            'repositoryId': self.repository_id,
            'name': self.name,
            'description': self.description,
            'exampleInput': self.example_input,
            'exampleOutput': self.example_output,
            'modelType': self.model_type.value,
            'modelServerless': self.model_serverless,
            'explainerType': self.explainer_type.value,
            'explainerServerless': self.explainer_serverless,
            'branchName': self.branch_name,
            'commit': self.commit,
        }



PYTORCH_CONFIG_FILE = """inference_address=http://0.0.0.0:8085
management_address=http://0.0.0.0:8081
metrics_address=http://0.0.0.0:8082
enable_metrics_api=true
metrics_format=prometheus
number_of_netty_threads=4
job_queue_size=10
service_envelope=kfserving
model_store=/mnt/models/model-store
model_snapshot={"name":"startup.cfg","modelCount":1,"models":{"model":{"1.0":{"defaultVersion":true,"marName":"model.mar","minWorkers":1,"maxWorkers":5,"batchSize":1,"maxBatchDelay":5000,"responseTimeout":120}}}}

"""


from typing import Any
from os.path import join

from tensorflow import Module

from . import BaseModel
from deeploy.enums import ModelType


class TensorFlowModel(BaseModel):

    __tensorflow_model: Module

    def __init__(self, model_object: Any, **kwargs) -> None:

        if not issubclass(type(model_object), Module):
            raise Exception('Not a valid TensorFlow class')

        self.__tensorflow_model = model_object
        return

    def save(self, local_folder_path: str) -> None:
        self.__tensorflow_model.save(join(local_folder_path, '1'))
        return

    def get_model_type(self) -> ModelType:
        return ModelType.TENSORFLOW


from typing import Any
from os.path import join

import dill
from shap.explainers.explainer import Explainer

from . import BaseExplainer
from deeploy.enums import ExplainerType


class SHAPExplainer(BaseExplainer):

    __shap_explainer: Explainer

    def __init__(self, explainer_object: Any) -> None:

        if not issubclass(type(explainer_object), Explainer):
            raise Exception('Not a valid SHAP class')

        self.__shap_explainer = explainer_object
        return

    def save(self, local_folder_path: str) -> None:
        with open(join(local_folder_path, 'explainer.dill'), 'wb') as f:
            dill.dump(self.__shap_explainer, f)
        return

    def get_explainer_type(self) -> ExplainerType:
        return ExplainerType.SHAP_KERNEL



import logging
from typing import Optional, List, Any, Callable
import os
import shutil
import uuid

from pydantic import BaseModel
import yaml
import requests

from .services import DeeployService, GitService, ModelWrapper, ExplainerWrapper
from .models import Repository, ClientConfig, Deployment, CreateDeployment, DeployOptions, V1Prediction, V2Prediction
from .enums import ExplainerType
from .common import delete_all_contents_in_directory, directory_exists, directory_empty


class Client(object):
    """ 
    A class for interacting with Deeploy
    """

    __config: ClientConfig

    def __init__(self, access_key: str, secret_key: str, host: str, workspace_id: str,
                 branch_name: str = None) -> None:
        """Initialise the Deeploy client
        Parameters:
            access_key (str): Personal Access Key generated from the Deeploy UI
            secret_key (str): Secret Access Key generated from the Deeploy UI
            host (str): The host at which Deeploy is located, i.e. deeploy.example.com
            workspace_id (str): The ID of the workspace in which your repository 
                is located
            branch_name (str, optional): The banchname on which to commit new models. 
                Defaults to the current branchname.
        """
        self.__config = ClientConfig(**{
            'access_key': access_key,
            'secret_key': secret_key,
            'host': host,
            'workspace_id': workspace_id,
            'branch_name': branch_name,
            'repository_id': '',
        })

        self.__deeploy_service = DeeployService(
            access_key,
            secret_key,
            host
        )

        if not self.__are_clientoptions_valid(self.__config):
            raise Exception('Client options not valid')

        return

    def deploy(self, model: Any, options: DeployOptions, local_repository_path: str, explainer: Any = None, overwrite: bool = False,
               commit_message: str = None) -> Deployment:
        """Deploy a model on Deeploy
        Parameters:
            model (Any): The class instance of an ML model
            options (DeployOptions): An instance of the deploy options class 
                containing the deployment options
            local_repository_path (str): Absolute path to the local git repository 
                which is connected to Deeploy
            explainer (Any, optional): The class instance of an optional model explainer
            overwrite (bool, optional): Whether or not to overwrite files that are in the 'model' and 'explainer' 
                folders in the git folder. Defaults to False
            commit_message (str, optional): Commit message to use
        """
        git_service = GitService(local_repository_path)

        repository_in_workspace, repository_id = self.__is_git_repository_in_workspace(git_service)

        if not repository_in_workspace:
            raise Exception(
                'Repository was not found in the Deeploy workspace. Make sure you have connected it before')
        else:
            self.__config.repository_id = repository_id

        logging.info('Pulling from the remote repository...')
        git_service.pull()
        logging.info('Successfully pulled from the remote repository.')

        logging.info('Saving the model to disk...')
        model_wrapper = ModelWrapper(model, 
            pytorch_model_file_path=options.pytorch_model_file_path, 
            pytorch_torchserve_handler_name=options.pytorch_torchserve_handler_name)
        self.__prepare_model_directory(git_service, local_repository_path, overwrite)
        model_folder = os.path.join(
            local_repository_path, 'model')
        model_wrapper.save(model_folder)
        total_file_sizes_model = self.__get_upload_size(model_folder)
        blob_storage_link = self.__upload_folder_to_blob(local_repository_path, model_folder)
        shutil.rmtree(model_folder)
        os.mkdir(model_folder)
        self.__create_reference_file(model_folder, blob_storage_link)
        git_service.add_folder_to_staging('model')
        commit_message = '[Deeploy Client] Add new model'

        if explainer:
            logging.info('Saving the explainer to disk...')
            explainer_wrapper = ExplainerWrapper(explainer)
            self.__prepare_explainer_directory(git_service, local_repository_path, overwrite)
            explainer_folder = os.path.join(
                local_repository_path, 'explainer')
            explainer_wrapper.save(explainer_folder)
            total_file_sizes_explainer = self.__get_upload_size(explainer_folder)
            blob_storage_link = self.__upload_folder_to_blob(local_repository_path, explainer_folder)
            shutil.rmtree(explainer_folder)
            os.mkdir(explainer_folder)
            self.__create_reference_file(explainer_folder, blob_storage_link)
            git_service.add_folder_to_staging('explainer')
            commit_message += ' and explainer'
            explainer_type = explainer_wrapper.get_explainer_type()
        else:
            explainer_type = ExplainerType.NO_EXPLAINER

        logging.info('Committing and pushing the result to the remote.')
        commit_sha = git_service.commit(commit_message)
        git_service.push()

        deployment_options = {
            'repository_id': self.__config.repository_id,
            'name': options.name,
            'description': options.description,
            # TODO: example input & output
            'example_input': options.example_input,
            'example_output': options.example_output,
            'model_type': model_wrapper.get_model_type().value,
            'model_serverless': options.model_serverless,
            'branch_name': git_service.get_current_branch_name(),
            'commit': commit_sha,
            'explainer_type': explainer_type.value,
            'explainer_serverless': options.explainer_serverless
        }

        deployment = self.__deeploy_service.create_deployment(
            self.__config.workspace_id, CreateDeployment(**deployment_options))

        return deployment

    def predict(self, deployment_id: str, request_body: dict) -> V1Prediction or V2Prediction:
        """Make a predict call
        Parameters:
            deployment_id (str): ID of the Deeploy deployment
            request_body (dict): Request body with input data for the model
        """

        workspace_id = self.__config.workspace_id
        prediction = self.__deeploy_service.predict(workspace_id, deployment_id, request_body)
        return prediction

    def explain(self, deployment_id: str, request_body: dict, image: bool = False) -> object:
        """Make an explain call
        Parameters:
            deployment_id (str): ID of the Deeploy deployment
            request_body (dict): Request body with input data for the model
            image (bool): Return image or not
        """
        workspace_id = self.__config.workspace_id
        explanation = self.__deeploy_service.explain(workspace_id, deployment_id, request_body, image)
        return explanation

    def __are_clientoptions_valid(self, config: ClientConfig) -> bool:
        """Check if the supplied options are valid
        """
        try:
            self.__deeploy_service.get_workspace(config.workspace_id)
        except Exception as e:
            raise e

        return True

    def __is_git_repository_in_workspace(self, git_service: GitService) -> (bool, str):
        remote_url = git_service.get_remote_url()
        workspace_id = self.__config.workspace_id

        repositories = self.__deeploy_service.get_repositories(workspace_id)
        correct_repositories = list(
            filter(lambda x: x.git_ssh_pull_link == remote_url, repositories))

        if len(correct_repositories) != 0:
            repository_id = correct_repositories[0].id
            return True, repository_id

        return False, None

    def __prepare_model_directory(self, git_service: GitService, local_repository_path: str, 
        overwrite=False) -> None:
        model_folder_path = os.path.join(
            local_repository_path, 'model')
        if not directory_exists(model_folder_path):
            try:
                os.mkdir(model_folder_path)
            except OSError:
                logging.error("Creation of the directory %s failed" %
                              model_folder_path)
        elif not directory_empty(model_folder_path):
            if not overwrite:
                raise Exception(
                    'The folder %s is not empty. Pass \'overwrite=True\' to overwrite contents.' % model_folder_path)
            delete_all_contents_in_directory(model_folder_path)
            git_service.delete_folder_from_staging('model')
        else:  # folder exists and empty
            pass
        return

    def __prepare_explainer_directory(self, git_service: GitService, local_repository_path: str, overwrite=False) -> None:
        explainer_folder_path = os.path.join(
            local_repository_path, 'explainer')
        if not directory_exists(explainer_folder_path):
            try:
                os.mkdir(explainer_folder_path)
            except OSError:
                logging.error("Creation of the directory %s failed" %
                              explainer_folder_path)
        elif not directory_empty(explainer_folder_path):
            if not overwrite:
                raise Exception(
                    'The folder %s is not empty. Pass \'overwrite=True\' to overwrite contents.' % explainer_folder_path)
            delete_all_contents_in_directory(explainer_folder_path)
            git_service.delete_folder_from_staging('explainer')
        else:  # folder exists and empty
            pass
        return

    def __get_upload_size(self, local_folder_path: str) -> int:
        total_file_sizes = 0
        for root, _, files in os.walk(local_folder_path):
            for single_file in files:
                file_path = os.path.join(root, single_file)
                file_size = os.path.getsize(file_path)
                total_file_sizes += file_size
        return total_file_sizes

    def __upload_folder_to_blob(self, local_repository_path: str, local_folder_path: str) -> str:
        upload_locations = list()
        blob_folder_uuid = str(uuid.uuid4())
        relative_folder_path = os.path.relpath(local_folder_path, local_repository_path)
        for root, _, files in os.walk(local_folder_path):
            for single_file in files:
                relative_file_path = os.path.join(relative_folder_path, os.path.relpath(root, local_folder_path))
                file_path = os.path.join(root, single_file)
                blob_file_location = self.__deeploy_service.upload_blob_file(file_path, relative_file_path, \
                    self.__config.workspace_id, self.__config.repository_id, blob_folder_uuid)
                upload_locations.append(blob_file_location)

        partition = upload_locations[0].partition(relative_folder_path)
        blob_folder_path = partition[0] + partition[1]
        return blob_folder_path
    
    def __create_reference_file(self, local_folder_path: str, blob_storage_link: str) -> None:
        file_path = os.path.join(local_folder_path, 'reference.yaml')
        data = { 'reference': blob_storage_link }

        with open(file_path, 'w') as outfile:
            yaml.dump(data, outfile, default_flow_style=False)
        return



from typing import Optional, List, Any

from pydantic import BaseModel


class ClientConfig(BaseModel):
    """
    Class containing the Deeploy client options

    Attributes:
      access_key: string representing the personal access key
      secret_key: string representing the personal secret key
      host: string representing the domain on which Deeploy is hosted
      workspace_id: string representing the workspace id in which to create
        deployments
      local_repository_path: string representing the relative or absolute path
        to the local git repository
      branch_name: string representing the branch name on which to commit. Defaults
        to the local active branch
    """
    access_key: str
    secret_key: str
    host: str
    workspace_id: str
    repository_id: str
    branch_name: Optional[str]



from typing import Optional

from git import Repo, Remote


class GitService(object):
    """ 
    A class for interacting with a local Git project
    """

    repository: Repo
    remote: Remote

    def __init__(self, local_repository_path: str, branch_name: str = None) -> None:
        """Initialise the Git client
        """
        # TODO: branch name
        self.repository = Repo(local_repository_path)
        self.remote = self.repository.remote('origin')

        if not self.__is_valid_git_project():
            raise Exception('Not a valid git project')

        return

    def __is_valid_git_project(self) -> bool:
        """Check if the supplied repository is valid
        """
        try:
            assert not self.repository.bare
        except:
            return False

        return True

    def add_folder_to_staging(self, relative_folder_path: str) -> None:
        """Add the folder and all its contents to the git staging area

        Parameters
        ----------
          relative_folder_path: str 
            represents the relative path to the folder from the root of 
            the git directory
        """
        self.repository.index.add([relative_folder_path])
        return

    def delete_folder_from_staging(self, relative_folder_path: str) -> None:
        self.repository.index.remove([relative_folder_path], False, r=True, ignore_unmatch=True)
        return

    def commit(self, commit_message: str) -> str:
        """Create a new commit on the current branch

        Parameters
        ----------
          commit_message: str
            representing the commit message
        """
        self.repository.index.commit(commit_message)
        return self.repository.head.commit.hexsha

    def pull(self) -> None:
        """Pull from the default remote repository
        """
        self.remote.pull(ff_only=True)
        return

    def push(self) -> None:
        """Push to the default remote repository
        """
        self.remote.push()
        return

    def get_remote_url(self) -> str:
        return self.remote.url

    def get_current_branch_name(self) -> str:
        return self.repository.active_branch.name



import setuptools

with open("docs/pypi.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='deeploy',
    version='0.1.0',
    description='The official Deeploy client for Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Lars Suanet',
    author_email='lars@deeploy.ml',
    packages=setuptools.find_packages(),
    url="https://gitlab.com/deeploy-ml/deeploy-python-client",
    project_urls={
        "Documentation": "https://deeploy-ml.gitlab.io/deeploy-python-client/",
        "Deeploy website": "https://deeploy.ml",
    },
    install_requires=[
        "pydantic>=1.7.3",
        "gitpython>=3.1.12",
        "requests>=2.25.1",
        "joblib>=1.0.1",
        "dill>=0.3.3",
        "ipython>=7.20.0",
        "nbconvert>=6.0.7",
        "pyyaml==5.4.1",
        "torch-model-archiver==0.3.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from typing import Any

from deeploy.enums import ModelType

class BaseModel:

    def __init__(self, model_object: Any, **kwargs) -> None:
        return

    def save(self, local_folder_path: str) -> None:
        return

    def get_model_type(self) -> ModelType:
        return
